"""methods for price signals"""

import pandas as pd
import numpy as np


def sma_signal(df: pd.DataFrame, colnames=["SMA_12", "SMA_26"]) -> pd.Series:
    def sma(serie: pd.Series, colnames: list) -> int or np.nan:
        """Simple Moving Average signal

        Args:
            serie (pd.Series): _description_
            colnames (list): _description_

        Returns:
            int or np.nan: _description_
        """
        if not pd.isna(serie[colnames[0]]) and not pd.isna(serie[colnames[1]]):
            if serie[colnames[0]] < serie[colnames[1]]:
                return 0
            else:
                return 1
        else:
            np.nan

    return df.apply(sma, args=(colnames,), axis=1)


def macd_signal(df: pd.DataFrame) -> pd.Series:
    def macd(serie: pd.Series) -> int or np.nan:
        """MACD signal

        Args:
            serie (pd.Series): _description_

        Returns:
            int or np.nan: _description_
        """
        if not pd.isna(serie["MACDh_12_26_9"]):
            if serie["MACDh_12_26_9"] > 0:
                return 1
            else:
                return 0
        else:
            return np.nan

    return df.copy().apply(macd, axis=1)


def bb_signal(df: pd.DataFrame, buffer=0.1) -> pd.Series:
    df["PREV_STOCK"] = df["Close"].shift(1)
    df["PREV_LOWERBB"] = df["BBL_20_2.0"].shift(1)
    df["PREV_UPPERBB"] = df["BBU_20_2.0"].shift(1)

    # IF PREV_STOCK > PREV_LOWERBB & CUR_STOCK < CUR_LOWER_BB => BUY
    # IF PREV_STOCK < PREV_UPPERBB & CUR_STOCK > CUR_UPPER_BB => SELL
    def bb(serie: pd.Series, buffer: float) -> int or np.nan:
        """Bollinger Bands signal

        Args:
            serie (pd.Series): _description_
            buffer (float): _description_

        Returns:
            int or np.nan: _description_
        """
        if (
            not pd.isna(serie["PREV_STOCK"])
            and not pd.isna(serie["PREV_LOWERBB"])
            and not pd.isna(serie["PREV_UPPERBB"])
        ):
            if (
                serie["PREV_STOCK"]
                > serie["PREV_LOWERBB"] + serie["PREV_LOWERBB"] * buffer
                and serie["Close"] < serie["BBL_20_2.0"] + serie["BBL_20_2.0"] * buffer
            ):
                return 1
            elif (
                serie["PREV_STOCK"]
                < serie["PREV_UPPERBB"] - serie["PREV_UPPERBB"] * buffer
                and serie["Close"] > serie["BBU_20_2.0"] - serie["BBU_20_2.0"] * buffer
            ):
                return 0
        else:
            return np.nan

    df["BBL_20_2_Signal"] = df.apply(bb, args=(buffer,), axis=1)
    df["BBL_20_2_Signal"] = df["BBL_20_2_Signal"].fillna(method="bfill")
    df["BBL_20_2_Signal"] = df["BBL_20_2_Signal"].fillna(method="ffill")
    return df["BBL_20_2_Signal"].copy()


def rsi_signal(df: pd.DataFrame, lower=30, upper=70) -> pd.Series:
    df["PREV_RSI"] = df["RSI_14"].shift(1)
    # IF PREVIOUS RSI > 30 AND CURRENT RSI < 30 ==> BUY SIGNAL
    # IF PREVIOUS RSI < 70 AND CURRENT RSI > 70 ==> SELL SIGNAL
    def rsi(serie: pd.Series, lower: int, upper: int) -> int or np.nan:
        """RSI signal

        Args:
            serie (pd.Series): _description_
            lower (int, optional): _description_. Defaults to 30.
            upper (int, optional): _description_. Defaults to 70.

        Returns:
            int or np.nan: _description_
        """
        if not pd.isna(serie["PREV_RSI"]):
            if serie["PREV_RSI"] > lower and serie["RSI_14"] < lower:
                return 1
            elif serie["PREV_RSI"] < upper and serie["RSI_14"] > upper:
                return 0
        else:
            return np.nan

    df["RSI_14_Signal"] = df.apply(
        rsi,
        args=(
            lower,
            upper,
        ),
        axis=1,
    )
    df["RSI_14_Signal"] = df["RSI_14_Signal"].fillna(method="bfill")
    df["RSI_14_Signal"] = df["RSI_14_Signal"].fillna(method="ffill")

    return df["RSI_14_Signal"].copy()



"""figures built for crypto prices."""

import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.colors import LinearSegmentedColormap, Normalize
from matplotlib.cm import ScalarMappable
import matplotlib.dates as mdates

import pandas as pd
import copy
import calplot
import pyfolio as pf
import dateutil
import datetime
import pytz
from dateutil.relativedelta import relativedelta

from surfingcrypto.ts import TS
from surfingcrypto.portfolio import Portfolio
from surfingcrypto.algotrading.backtesting import BackTest
import surfingcrypto.reporting.plotting as scplot
from surfingcrypto.reporting.plotting import shiftedColorMap


# # GLOBAL VARIABLES (?!?) TO SET PLOT STYLE
plt.style.use("dark_background")
mpl.rcParams["font.size"] = 4


class BaseFigure:
    """Base class object for all figures.

    This class contains a set of common methods useful
    for the generic plots required by the `surfingcrypto`
    package.

    Note:
        The `BaseFigure` class does not create a `figure`
        nor any `ax`. This is done by children classes.

    Arguments:
        object (:obj:`surfingcrypto.ts.TS` or :obj:`surfingcrypto.portfolio.Portfolio`):
            timeseries or portfolio object
        graphstart (str) : date string in dd-mm-yyyy format
            (or relative from today eg. 1 month: `1m`,3 month: `3m`) from which
            to start the graph.
    
    Attributes:
        object (:obj:`surfingcrypto.ts.TS` or :obj:`surfingcrypto.portfolio.Portfolio`):
            timeseries or portfolio object
        graphstart (:obj:`datetime.date`) : start of fraph, UTC aware
        graphend (:obj:`datetime.date` or `None`): end of graph, UTC aware
    
    """

    def __init__(
        self,
        object: TS or Portfolio,
        graphstart="1-1-2021",
    ):
        self.object = copy.copy(object)  # copy so can be sliced for purpose
        self._set_graphstart(graphstart)
        self.graphend = None # not implemented yet
        # if ts, slice df to extent
        if isinstance(object, TS):
            self.object.df = self.object.df.loc[self.graphstart:self.graphend]

    def _set_graphstart(self, graphstart):
        if graphstart.lower() == "3m":
            self.graphstart = datetime.date.today() + relativedelta(months=-3)
        elif graphstart.lower() == "6m":
            self.graphstart = datetime.date.today() + relativedelta(months=-6)
        elif graphstart.lower() == "1m":
            self.graphstart = datetime.date.today() + relativedelta(months=-1)
        elif graphstart.lower() == "1y":
            self.graphstart = datetime.date.today() + relativedelta(years=-1)
        else:
            self.graphstart = dateutil.parser.parse(graphstart).date()
        # localize UTC
        self.graphstart = datetime.datetime.combine(
            self.graphstart, datetime.time.min
        ).replace(tzinfo=pytz.UTC)

    def save(self, path):
        """
        Save fig to specified path.

        Arguments:
            path (str) : path to output file.
        """
        return self.f.savefig(path)

    def set_axes(self):
        """Set axes look.

        Set axes look.
        Set axes to the specified xlims.
        """
        if hasattr(self, "axes"):
            for iax in self.axes:
                iax.grid(which="major", axis="x", linewidth=0.1)
                iax.grid(which="major", axis="y", linewidth=0.05)
                iax.yaxis.set_label_position("left")
                iax.yaxis.tick_left()
                iax.xaxis.set_major_locator(mdates.MonthLocator(bymonthday=1))
                iax.xaxis.set_minor_locator(mdates.WeekdayLocator(byweekday=mdates.MO))
                iax.xaxis.set_major_formatter(mdates.DateFormatter('%b %Y'))

        else: # if not axes, than ax
            self.ax.grid(which="major", axis="x", linewidth=0.1)
            self.ax.grid(which="major", axis="y", linewidth=0.05)
            self.ax.yaxis.set_label_position("left")
            self.ax.yaxis.tick_left()
            self.ax.xaxis.set_major_locator(mdates.MonthLocator(bymonthday=1))
            self.ax.xaxis.set_minor_locator(mdates.WeekdayLocator(byweekday=mdates.MO))
            self.ax.xaxis.set_major_formatter(mdates.DateFormatter('%b %Y'))


class SimplePlot(BaseFigure):
    """Simple plot.

    Candlesticks + volume.

    Arguments:
        ts (:class:`surfingcrypto.ts.TS`) : `surfingcrypto.ts.TS` object
        graphstart (str) : date string in d-m-Y format
            (or relative from today eg. 1 month: `1m`,3 month: `3m`)
            from which to start the graph.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_plot()

    def default_plot(self):
        """Plot default plot."""
        # figure
        self.f, self.axes = plt.subplots(
            2,
            1,
            sharex=True,
            gridspec_kw={"height_ratios": [4, 1]},
            dpi=200,
            figsize=(9, 3),
        )
        # plot
        scplot.candlesticks(
            self.object, # a TS object
            ax=self.axes[0],
            volume=True,
            vol_ax=self.axes[1],
            style="candlesticks",
        )
        # axes look
        self.set_axes()
        self.axes[0].set_title(
            self.object.coin, fontsize=10, va="center", ha="center", pad=20
        )
        # log
        print(f"{self.object.coin} plotted.")


class TaPlot(BaseFigure):
    """Ta plot.

    This is the Technical Analysis plot.
    It shows (at current time): candlesticks,
    volume and 3 TA Indicators (MACD, BB bands and RSI)
    Can be easily modified to fit other and/or more indicators.

    Arguments:
        ts (:class:`surfingcrypto.ts.TS`) : `surfingcrypto.ts.TS` object
        graphstart (str) : date string in d-m-Y format
            (or relative from today eg. 1 month: `1m`,3 month: `3m`) from
            which to start the graph.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ta_plot()

    def ta_plot(self):
        """Plot ta indicators."""
        if not hasattr(self.object, "ta_params"):
            raise AttributeError("Object myust have `ta_params`")
        if not all(
            elem in list(self.object.ta_params.keys())
            for elem in ["macd", "sma", "bbands", "rsi"]
        ):
            raise AttributeError(
                "Object must have `sma`,`macd`,`bbands` "
                "and `rsi` in `ta_params` attribute."
            )
        # figure
        self.f, self.axes = plt.subplots(
            5,
            1,
            sharex=True,
            gridspec_kw={"height_ratios": [2, 1, 1, 1, 1]},
            dpi=200,
            figsize=(7.5, 7.5),
        )
        # plots
        scplot.candlesticks(
            self.object,
            ax=self.axes[0],
            volume=True,
            vol_ax=self.axes[1],
            style="candlesticks",
        )
        scplot.plot_moving_averages(self.object, ax=self.axes[0])
        scplot.plot_macd(self.object, self.axes[2], plot_lines=False)
        scplot.candlesticks(self.object, self.axes[3], style="ohlc")
        scplot.plot_bb(self.object, self.axes[3])
        scplot.plot_RSI(self.object, self.axes[4])

        # axes look
        self.set_axes()
        self.axes[0].set_title(
            self.object.coin, fontsize=10, va="center", ha="center", pad=20
        )

        ymin = self.object.df["Low"].min() - 0.1 * self.object.df["Low"].min()
        ymax = self.object.df["High"].max() + 0.1 * self.object.df["High"].max()
        self.axes[0].set_ylim([ymin, ymax])

        # log
        print(f"{self.object.coin} plotted.")


class ATHPlot(BaseFigure):
    """
    distance from ATH plot.

    Arguments:
        ts (:class:`surfingcrypto.ts.TS`) : `surfingcrypto.ts.TS` object
        graphstart (str) : date string in d-m-Y format
            (or relative from today eg. 1 month: `1m`,3 month: `3m`)
            from which to start the graph.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.plot()

    def plot(self):
        """
        Plot distance from ath points.

        Note:
            ATM it its a zoomed-in view.
        """
        # figure
        self.f, self.ax = plt.subplots(dpi=200, figsize=(9, 3))
        # compute distance

        # normalizzato su tutto intervallo
        vmin = self.object.df["distance_ATH"].min()
        vmax = self.object.df["distance_ATH"].max()
        norm = Normalize(vmin=vmin, vmax=vmax)

        cmap = LinearSegmentedColormap.from_list(
            "colorbar",
            [
                "green",
                "orange",
                "red",
                "magenta",
            ],
        )
        colors = [cmap(norm(x)) for x in self.object.df["distance_ATH"]]

        # points
        self.ax.scatter(self.object.df.index, self.object.df.Close, c=colors, s=2)
        # colorbar
        cmappable = ScalarMappable(norm=norm, cmap=cmap)
        self.f.colorbar(cmappable, ax=self.ax)

        # axes look
        self.set_axes()
        self.ax.set_title(
            "Distance from All Time High: " + self.object.coin,
            fontsize=10,
            va="center",
            ha="center",
            pad=20,
        )
        # log
        print(f"{self.object.coin} ATH plotted.")


class PortfolioPlot(BaseFigure):
    """Portfolio plots.

    This is the Portfolio plot.

    Arguments:
        variables (list): list of variables to plot
        by_symbol (bool, optional): _description_. Defaults to False.
        zero_line (bool, optional): _description_. Defaults to False.
        portfolio (:class:`surfingcrypto.portfolio.portfolio`) :
            `surfingcrypto.portfolio.portfolio` object
        graphstart (str) : date string in d-m-Y format
            (or relative from today eg. 1 month: `1m`,3 month: `3m`)
            from which to start the graph.

    """

    def __init__(
        self,
        variables: list,
        by_symbol=False,
        zero_line=False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.plot(variables, by_symbol, zero_line)

    def plot(self, variables: list, by_symbol: bool, zero_line: bool):
        """Plot portfolio plot."""
        # figure
        self.f, self.ax = plt.subplots(
            dpi=200,
        )
        df = self.object.tracker.daily_grouped_metrics(variables, by_symbol=by_symbol)
        df = df.loc[self.graphstart.date() :].dropna(axis=1, how="all")
        df.plot(ax=self.ax)
        if zero_line:
            self.ax.axhline(0, color="r")

        self.ax.legend(bbox_to_anchor=(1, 1))


class CalendarPlot:
    """Calendar plot.

    This object is a good visualisation of a variable in a calendar style.

    Args:
        values (pd.Series): variable to plot
    """

    def __init__(self, values: pd.Series):
        norm = Normalize(
            vmin=values["Stock Gain / (Loss)"].min(),
            vmax=values["Stock Gain / (Loss)"].max(),
        )
        cmap = shiftedColorMap(
            LinearSegmentedColormap.from_list(
                "colorbar",
                [
                    "darkred",
                    "red",
                    "orange",
                    "grey",
                    "lightgreen",
                    "green",
                    "darkgreen",
                ],
            ),
            midpoint=norm(0),
        )
        self.calplot = calplot.calplot(values["Stock Gain / (Loss)"], cmap=cmap)


class BacktestPerformancePlot:
    """Bakctest performance plot."""

    def __init__(self, bt: BackTest):
        self.bt = bt
        self.plot()

    def plot(self):
        """Plot backtest performance."""
        self.fig, axes = plt.subplots(
            nrows=2, ncols=2, figsize=(16, 9), constrained_layout=True
        )
        self.axes = axes.flatten()

        pf.plot_drawdown_periods(returns=self.bt.returns, ax=self.axes[0])
        self.axes[0].grid(True)
        pf.plot_rolling_returns(
            returns=self.bt.returns,
            factor_returns=self.bt.benchmark_returns,
            ax=self.axes[1],
            title="Strategy vs Buy&Hold",
        )
        self.axes[1].grid(True)
        pf.plot_drawdown_underwater(returns=self.bt.returns, ax=self.axes[2])
        self.axes[2].grid(True)
        pf.plot_rolling_sharpe(returns=self.bt.returns, ax=self.axes[3])
        self.axes[3].grid(True)



"""model object"""

import pandas as pd
import copy

from sklearn import linear_model
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from surfingcrypto.algotrading.features import Features


class Model:
    def __init__(self, name: str, f: Features):
        self.name = name
        self.feature = copy.copy(f)
        self.model = self._set_model()
        self.X = self.feature.model_df[self.feature.x_cols_names].copy()
        self.Y = self.feature.model_df["direction"].copy()
        self._fit_model()
        self.estimated = self._estimate()

    def _set_model(self):
        if self.name == "svm":
            return SVC()
        elif self.name == "log_reg":
            return linear_model.LogisticRegression()
        elif self.name == "gauss_nb":
            return GaussianNB()
        elif self.name == "random_forest":
            return RandomForestClassifier(max_depth=10, n_estimators=100)
        elif self.name == "MLP":
            return MLPClassifier(max_iter=500)
        else:
            raise NotImplementedError

    def _fit_model(self):
        self.model.fit(self.X, self.Y)

    def _estimate(self) -> pd.Series:
        return pd.DataFrame(
            self.model.predict(self.X),
            index=self.X.index,
            columns=["predicted"],
        )

    def make_tomorrow_prediction(self) -> int:
        return self.model.predict(
            self.feature.get_future_x().to_numpy().reshape(1, -1)
        )[0]

    def __repr__(self) -> str:
        return f"Model(name={self.name})"

    def __str__(self) -> str:
        return f"Model(name={self.name})"



"""feature engineering"""

import pandas as pd
import numpy as np
import re

from surfingcrypto.ts import TS
import surfingcrypto.algotrading.signals as surf_signals


class Features:
    def __init__(self, ts: TS, indicators: list):
        self.ts = ts
        self.indicators = self._fmt_indicators(indicators)
        self.x_cols_names = []
        self.df = pd.DataFrame()  # dataframe with  indicators and signals
        self.model_df = pd.DataFrame()  # dataframe with Y and X

    def _fmt_indicators(self, indicators: list) -> dict:
        dicts = {}
        for i in range(len(indicators)):
            dicts["i_" + "{}".format(str(i + 1).zfill(2))] = indicators[i]
        return dicts


class BinaryLaggedLogReturns(Features):
    def __init__(self, lags: list, *args, **kwargs):
        self.lags = lags
        super().__init__(*args, **kwargs)
        self.df = self._compute_signals()
        self.model_df = self._set_binary_lagged_features(self.indicators)

    def _compute_signals(self) -> pd.DataFrame:
        df = self.ts.df

        # SMA
        df["SMA_12_26_Signal"] = surf_signals.sma_signal(df)
        df["SMA_100_200_Signal"] = surf_signals.sma_signal(
            df, colnames=["SMA_100", "SMA_200"]
        )
        # MACD
        df["MACD_12_26_9_Signal"] = surf_signals.macd_signal(df)

        ## BB
        df["BB_20_2_Signal"] = surf_signals.bb_signal(df)

        # RSI
        df["RSI_14_Signal"] = surf_signals.rsi_signal(df)

        return df

    def _set_binary_lagged_features(self, indicatori: dict) -> pd.DataFrame:
        model_df = self.df[["Close"] + list(indicatori.values())].copy()

        model_df.rename(columns={"Close": self.ts.coin}, inplace=True)
        model_df.rename(
            columns={v: k for k, v in indicatori.items()}, inplace=True
        )  # rename cols with reversed dict

        # calculate daily log returns
        model_df["returns"] = np.log(
            model_df[self.ts.coin] / model_df[self.ts.coin].shift(1)
        )
        model_df.dropna(inplace=True)

        # calculate binary direction of market
        model_df["direction"] = np.sign(model_df["returns"]).astype(int)

        # get lagged values of all indicators
        self.x_cols_names = []
        for key in indicatori:
            for lag in self.lags:
                col = f"{key}_lag{str(lag).zfill(2)}"
                model_df[col] = model_df[key].shift(lag)
                self.x_cols_names.append(col)

        model_df.dropna(inplace=True)

        return model_df

    def get_future_x(self) -> pd.Series:
        last = self.model_df.loc[self.model_df.iloc[-1].name, self.x_cols_names]
        future = []
        for key in self.indicators:
            iseries = last.loc[last.index.str.contains(key)].shift()
            iseries.iloc[0] = self.model_df.loc[self.model_df.iloc[-1].name, key]
            future.append(iseries)

        future = pd.concat(future)
        future.name = self.model_df.index[-1] + pd.Timedelta(days=1)

        return future

    def __repr__(self) -> str:
        return f"BinaryLaggedLogReturns(ts={self.ts.coin},lags={self.lags})"

    def __str__(self) -> str:
        return f"BinaryLaggedLogReturns(ts={self.ts.coin}),lags={self.lags})"



"""
time-series objects for cryptocurrencies.
"""
import pandas as pd
import os
import pandas_ta as ta
from surfingcrypto.config import Config


class TS:
    """
    This is an time-series oriented crypto price data object.

    Note:
        Data can be downloaded calling ```surfingcrypto.scraper.Scraper```
        object.

    Arguments:
        config (:obj:`surfingcrypto.config.config`): configuration
            object
        coin (str): string representing the crypto coin of choice,
            eg. BTC,ETH

    Attributes:
        config (:obj:`surfingcrypto.config.config`): package
            configuration object
        coin (str): string representing the crypto coin of choice,
            eg. BTC,ETH
        fiat (str): fiat currency in which the prices are represented
        df (:obj:`pandas.DataFrame`): dataframe with datetime index
            of ohlc data. Could store also TA indicators if these are
            computed invoking the relative method.
        ta_params (dict): dictionary containing TA parametrization
    """

    def __init__(self, configuration: Config, coin: str):

        self.config = configuration
        self.fiat = self.config.fiat

        # default ta params
        self.ta_params = {
            "sma": [{"fast": 12, "slow": 26}, {"fast": 100, "slow": 200}],
            "macd": {"fast": 12, "slow": 26, "signal": 9},
            "bbands": {"length": 20, "std": 2},
            "rsi": {"timeperiod": 14},
        }
        # rebrandings
        if coin in self.config.rebrandings:
            coin = self.config.rebrandings[coin]

        self.coin = coin

        self._build_ts()

    def _build_ts(self):
        """
        reads the data from data stored locally in `data/ts/` and
        saved in .csv format.
        """
        if os.path.isfile(
            self.config.data_folder / "ts" / (self.coin + "_" + self.fiat + ".csv")
        ):
            self.df = pd.read_csv(
                self.config.data_folder / "ts" / (self.coin + "_" + self.fiat + ".csv")
            )
            self.df["Date"] = pd.to_datetime(self.df["Date"], utc=True)
            self.df.set_index("Date", inplace=True)
            self._validity_checks()
        else:
            raise FileNotFoundError(f"{self.coin}.csv not found.")

    def _validity_checks(self):
        """
        validity check to avoid errors later on

        Raises:
            AttributeError: _description_
            ValueError: _description_
        """
        # minimum columns
        if not set(
            [
                "Open",
                "High",
                "Low",
                "Close",
            ]
        ).issubset(self.df.columns):
            raise AttributeError(
                "df must have at least columns named: " "Open, High, Low, Close"
            )
        # duplicates
        if any(self.df.index.duplicated()):
            raise ValueError("Data has duplicates.")

    def set_ta_params(self, params: dict):
        """
        sets new TA parameters::

            self.ta_params = {
                "sma": [
                    {"fast": 12, "slow": 26},
                    {"fast": 100, "slow": 200}
                    ],
                "macd": {"fast": 12, "slow": 26, "signal": 9},
                "bbands": {"length": 20, "std": 2},
                "rsi": {"timeperiod": 14},}

        """
        for param in params:
            # dictionary
            self.ta_params[param] = params[param]

    # COLUMNS FOR INDICATORS SAVED TO THE DF INPLACE

    def compute_ta_indicators(self):
        """computes the selected TA indicators and appends them to df attribute
        using `pandas_ta` module.

        It supports:
            - SMA
            - MACD
            - Bolinger bands
            - RSI
        """
        for key in self.ta_params:
            # sma can also be a list,
            if key == "sma":
                # all sma
                for sma in self.ta_params["sma"]:
                    self.df.ta.sma(length=sma["slow"], append=True)
                    self.df.ta.sma(length=sma["fast"], append=True)
            elif key == "macd":
                # macd
                self.df.ta.macd(
                    window_slow=self.ta_params["macd"]["slow"],
                    window_fast=self.ta_params["macd"]["fast"],
                    window_sign=self.ta_params["macd"]["signal"],
                    append=True,
                )
            elif key == "bbands":
                # bollinger bands
                self.df.ta.bbands(
                    length=self.ta_params["bbands"]["length"],
                    std=self.ta_params["bbands"]["std"],
                    append=True,
                )
            elif key == "rsi":
                self.df.ta.rsi(
                    timeperiod=self.ta_params["rsi"]["timeperiod"], append=True
                )
            else:
                raise NotImplementedError

    def distance_from_ath(self):
        d = {}
        for idx in self.df.index:
            d[idx] = abs(self.df.loc[idx, "Close"] - self.df[:idx]["Close"].max())
        self.df["distance_ATH"] = pd.Series(d)



"""
scraping price data from the internet.
"""
from cryptocmd import CmcScraper
import datetime
import pandas as pd
import os
import traceback
import pathlib


class Scraper:
    """
    gets all data required by configuration.

    #. Iterates through all coins specified in configuration.
    #. Starts and UpdateHandler object to identify the kind
        of update that is required.
    #. Uses selected apiwrapper for downloading data.

    Arguments:
        config (:obj:`surfingcrypto.config.config`): package
            configuration object

    Attributes:
        config (:obj:`surfingcrypto.config.config`): package
            configuration object
        runs (:obj:`list` of :obj:`surfincrypto.scraper.UpdateHandler`): list
            of `UpdateHandler` objects executed
        errors (:obj:`list` of :obj:`surfincrypto.scraper.UpdateHandler`): list
            of `UpdateHandler` objects that have errors
        output (bool): Overall boolean output of process, aka if
            everything went well.
        output_verbose (str): Overall string description of output.


    """

    def __init__(self, config):
        self.config = config
        self.runs = []
        self.errors = []
        self.output_verbose = ""
        self.output = None

    def run(self):
        """runs the scraping process for all coins."""
        for key in self.config.scraping_req:
            # dates are utc unaware
            start = self.config.scraping_req[key]["start"].date()
            end_day = self.config.scraping_req[key]["end_day"].date()

            # rebrandings
            if key in self.config.rebrandings:
                key = self.config.rebrandings[key]

            path = (
                self.config.data_folder / "ts" / (key + "_" + self.config.fiat + ".csv")
            )

            c = UpdateHandler(key, self.config.fiat, start, end_day, path)

            self.runs.append(c)

        self._log()

    def _log(self):
        """
        produce a log of all executions.
        """
        for run in self.runs:
            if run.error is not None:
                self.errors.append(run)

        if len(self.errors) == 0:
            self.output_verbose = "Update successful."
            self.output = True
        else:
            self.output_verbose = (
                "Update failed."
                f" There are ({len(self.errors)}/{len(self.runs)}) errors."
            )
            self.output = False


class UpdateHandler:
    """
    Checks if the file exists, if not it sets the date boundaries
    for the price API wrappers exactly as passed with the
    `start` and `end_day` arguments.
    If local data already exists, it checks first and last element of datetime column
    "Date" and compares with update request.

    Arguments:
        coin (str): symbol of crypto
        fiat (str): fiat of prices
        start (:obj:`datetime.datetime`):  start day
            from which is requested to scrape data
        end_day (:obj:`datetime.datetime`): end day of
            data requested, included
        path (:obj:`pathlib.Path`): path to csv file

    Attributes:
        coin (str): symbol of crypto
        fiat (str): fiat of prices
        start (:obj:`datetime.datetime`):  start day
            from which is requested to scrape data
        end_day (:obj:`datetime.datetime`): end day of
            data requested, included
        path (:obj:`pathlib.Path`): path to csv file
        left (:obj:`datetime.datetime` or :obj:`list` of `datetime.datetime`):
            left boundary (or boundaries, in case updating both sides of ts)
        right (:obj:`datetime.datetime` or :obj:`list` of `datetime.datetime`):
            right boundary (or boundaries, in case updating both sides of ts)
        df (:obj:`pandas.DataFrame`) : scraped df, SORTED IN ? ORDER
        description (str): verbose description of result
        result (bool): outcome of scraping run
        error (:obj:`Excection`): generic error
    """

    def __init__(
        self,
        coin: str,
        fiat: str,
        start: datetime.datetime,
        end_day: datetime.datetime,
        path: pathlib.Path,
        apiwrapper="cmc",
    ):
        self.coin = coin
        self.fiat = fiat
        self.apiwrapper = apiwrapper

        # dates are utc unaware
        self.start = start
        self.end_day = end_day

        self.path = path

        # log
        self.description = None
        self.result = None
        self.error = None

        if apiwrapper == "cmc":
            self.apiwrapper = CMCutility
        else:
            raise NotImplementedError

        self._handle_update()

    def _handle_update(self) -> None:
        """
        gets the bounds of the dataframes to be downloaded
        to fullfill update.
        """
        # check if exists
        if os.path.isfile(self.path):
            df, first, last = self._load_csv()

            if (first <= self.start) and (self.end_day <= last):
                self.df = df
                self.description = f"{self.coin} in {self.fiat}, already up to date."
                self.result = True
            else:
                try:
                    self.left, self.right = self._get_required_bounds(first, last)
                    self.df = self._get_updates(df)
                    self.df.to_csv(self.path, index=False)
                    self.description = (
                        f"{self.coin} in {self.fiat}, successfully updated."
                    )
                    self.result = True
                except:
                    self.description = f"{self.coin} in {self.fiat}," " update failed."
                    self.result = False
                    self.error = traceback.format_exc()
        else:
            try:
                self.left, self.right = self.start, self.end_day
                self.df = self._get_updates(None)
                self.df.to_csv(self.path, index=False)
                self.description = (
                    f"{self.coin} in {self.fiat}, successfully downloaded."
                )
                self.result = True
            except:
                self.description = f"{self.coin} in {self.fiat}, download failed."
                self.result = False
                self.error = traceback.format_exc()

    def _load_csv(self) -> tuple:
        """
        load local csv data from directory specified as `data_folder`
        in package config

        Return:
            df (:obj:`pandas.DataFrame`): dataframe of locally stored data
            first (:obj:`datetime.datetime`): first known price
            last (:obj:`datetime.datetime`): last known price
        """
        df = pd.read_csv(self.path)
        df["Date"] = pd.to_datetime(df["Date"])
        first = df["Date"].iloc[0]
        last = df["Date"].iloc[-1]
        return df, first, last

    def _get_required_bounds(
        self, first: datetime.datetime, last: datetime.datetime
    ) -> tuple:
        """
        finds info for the missing data in the local dataframe - represented
        by the `first` and `last` arguments - with respect to the
        request stored in `start` and `end_day` attribute

        Arguments:
            first (:obj:`datetime.datetime`): first known price
            last (:obj:`datetime.datetime`): last known price

        Raises:
            NotImplementedError: _description_

        Returns:
            tuple : 2-tuple of datetime or 2 lists of datetime,
                in case an update of both the front end and
                the last end is required.
        """
        # appending to front can cause exceptions because
        # coin data is available from a date that is later
        # than the start parameter

        # only to the front
        if self.start < first and last == self.end_day:
            left, right = self.start, (first - datetime.timedelta(1))

        # only to the end
        elif (first == self.start or first < self.start) and last < self.end_day:
            left, right = last + datetime.timedelta(1), self.end_day

        # both sides
        elif self.start < first and last < self.end_day:
            left, right = [], []
            # front part
            left.append(self.start)
            right.append(first - datetime.timedelta(1))
            # end part
            left.append(last + datetime.timedelta(1))
            right.append(self.end_day)

        else:
            print(f"first: {first}, last: {last}")
            print(f"start: {self.start}, end: {self.end_day}")
            raise NotImplementedError

        return left, right

    def _get_updates(self, df: pd.DataFrame or None) -> pd.DataFrame:
        """returns dataframe of prices

        get updates using selected apiwrapper.

        The dataframe has default integer index.

        Args:
            df (pd.DataFrame or None): _description_

        Raises:
            NotImplementedError: _description_

        Returns:
            pd.DataFrame: _description_
        """
        updates = [
            df,
        ]
        # one side update
        if isinstance(self.left, datetime.date) and isinstance(
            self.right, datetime.date
        ):
            updates.append(
                self.apiwrapper(
                    self.coin, self.left, self.right, self.fiat
                ).scrape_data()
            )
        # two side update (lists)
        else:
            for l, r in zip(self.left, self.right):
                updates.append(
                    self.apiwrapper(self.coin, l, r, self.fiat).scrape_data()
                )

        # concat
        df = pd.concat(updates)
        # sort ascending, oldest to newest
        df.sort_values(by="Date", inplace=True)
        df.reset_index(inplace=True, drop=True)
        return df

    def __str__(self) -> str:
        errors = hasattr(self, "error")
        return f"UpdateHandler({self.coin}-{self.fiat}: {self.description})"

    def __repr__(self) -> str:
        errors = hasattr(self, "error")
        return f"UpdateHandler({self.coin}-{self.fiat}: {self.description})"


class CMCutility(CmcScraper):
    """
    wrapper of GitHub repo `cryptocmd`.

    Scrapes data from coinmarketcap.com.
    """

    def __init__(
        self,
        coin: str,
        left: datetime.datetime,
        right: datetime.datetime,
        fiat: str,
    ):
        self.coin = coin
        self.left = left.strftime("%d-%m-%Y")
        self.right = right.strftime("%d-%m-%Y")
        self.fiat = fiat
        self.response = None
        super().__init__(self.coin, self.left, self.right, fiat=self.fiat)

    def scrape_data(self) -> pd.DataFrame:
        try:
            scraped = self.get_dataframe()
            self.response = True
        except:
            scraped = pd.DataFrame(columns=["Date"])
            self.response = False
        return scraped

    def __str__(self):
        return (
            f"CmcScraper({self.coin}"
            f", left={self.left}"
            f", right={self.right}"
            f", response={self.response}"
            ")"
        )

    def __repr__(self):
        return (
            f"CmcScraper({self.coin}"
            f", left={self.left}"
            f", right={self.right}"
            f", response={self.response}"
            ")"
        )



"""custom objects for backtesting"""

import backtrader as bt
import pandas as pd
import pyfolio as pf  # install with pip install git+ssh://git@github.com/giocaizzi/pyfolio.git
import matplotlib.pyplot as plt

from surfingcrypto.algotrading.model import Model


# class to define the columns we will provide
class CryptoPandasData(bt.feeds.PandasData):
    """pandas df structure"""

    cols = ["open", "high", "low", "close", "volume"] + ["predicted"]

    # create lines
    lines = tuple(cols)

    # define parameters
    params = {c: -1 for c in cols}
    params.update({"datetime": None})
    params = tuple(params.items())


class CryptoComissionInfo(bt.CommissionInfo):
    """commission info for crypto"""

    params = (
        ("commission", 1),  # percentage
        ("mult", 1.0),
        ("margin", None),
        ("commtype", None),
        ("stocklike", False),
        ("percabs", False),
        ("interest", 0.0),
        ("interest_long", False),
        ("leverage", 1.0),
        ("automargin", False),
    )

    def getsize(self, price, cash):
        """Returns fractional size for cash operation @price"""
        return self.p.leverage * (cash / price)


class CryptoSizer(bt.Sizer):
    """sizer for getting proper sizing"""

    params = (("prop", 1),)

    def _getsizing(self, comminfo, cash, data, isbuy):
        """Returns the proper sizing"""

        if isbuy:  # Buying
            target = (
                self.broker.getvalue() * self.params.prop
            )  # Ideal total value of the position
            price = data.close[0]
            size_net = target / price  # How many shares are needed to get target
            size = size_net * 0.99

            if size * price > cash:
                return 0  # Not enough money for this trade
            else:
                return size

        else:  # Selling
            return self.broker.getposition(data).size  # Clear the position


class BackTest:
    """backtest istance"""

    def __init__(
        self,
        m: Model,
        start: str,
        verbose=False,
    ):
        self.model = m
        self.start = start
        self.name = self.model.feature.ts.coin
        self.verbose = verbose

        # benchmark buy&hold returns
        self.benchmark_returns = self._get_benchmark_returns()

        # backtrader data
        data = self._fmt_dataframe()

        # instantiate Cerebro, add strategy, data, initial cash, commission and pyfolio for performance analysis
        self.cerebro = bt.Cerebro(stdstats=True, cheat_on_open=False)
        # cerebro.broker = bt.brokers.BackBroker(slip_open=True)  # consider market slippage
        self.cerebro.broker.setcash(1250.0)  #  cash value
        self.cerebro.adddata(CryptoPandasData(dataname=data), name=self.name)
        self.cerebro.addsizer(CryptoSizer)
        self.cerebro.addstrategy(MLStrategy, verbose=self.verbose)  # strategy
        self.cerebro.broker.addcommissioninfo(
            CryptoComissionInfo(),
        )  # fractional prices and commissions scheme
        self.cerebro.addanalyzer(bt.analyzers.PyFolio, _name="pyfolio")  # analyizer

    def _fmt_dataframe(self) -> pd.DataFrame:
        prices = self.model.feature.df[["Open", "High", "Low", "Close", "Volume"]]
        prices = prices.loc[self.start :]
        prices.rename(
            columns={
                "Open": "open",
                "High": "high",
                "Low": "low",
                "Close": "close",
                "Volume": "volume",
            },
            inplace=True,
        )

        backtestdata = self.model.estimated.join(prices, how="right").dropna()

        return backtestdata

    def run(self):
        # run the backtest
        self.start_value = self.cerebro.broker.getvalue()
        if self.verbose:
            print("Starting Portfolio Value: %.2f" % self.start_value)
        self.backtest_result = self.cerebro.run()
        self.returns = self._get_backtest_returns()
        self.end_value = self.cerebro.broker.getvalue()
        if self.verbose:
            print("Final Portfolio Value: %.2f" % self.end_value)

    def _get_benchmark_returns(self) -> pd.Series:
        # get benchmark returns # just buy and hold
        benchmark_rets = self.model.feature.model_df.loc[self.start :, "returns"]
        benchmark_rets.name = "Buy&Hold"
        return benchmark_rets

    def _get_backtest_returns(self) -> pd.Series:
        pyfoliozer = self.backtest_result[0].analyzers.getbyname("pyfolio")
        (
            returns,
            positions,
            transactions,
            gross_lev,
        ) = pyfoliozer.get_pf_items()
        returns.name = "Backtest returns"

        return returns

    def performance_stats(self) -> pd.Series:
        # Extract inputs for pyfolio
        performance = pf.timeseries.perf_stats(self.returns)
        performance = performance.append(
            pd.Series([self.end_value], index=["End Value"])
        )

        return performance

    def print_log(self):
        if len(self.backtest_result) == 1:
            print(self.backtest_result[0].log_text)
        else:
            raise NotImplementedError


# define backtesting strategy class
class MLStrategy(bt.Strategy):
    params = (("verbose", False),)

    def __init__(self):

        self.log_text = ""

        # keep track of open, close prices and predicted value in the series
        self.data_predicted = self.datas[0].predicted
        self.data_open = self.datas[0].open
        self.data_close = self.datas[0].close

        # keep track of pending orders/buy price/buy commission
        self.order = None
        self.price = None
        self.comm = None

    # logging function
    def log(self, txt):
        """Logging function"""
        dt = self.datas[0].datetime.date(0).isoformat()
        s = f"{dt}, {txt}"
        if self.params.verbose:
            print(s)
        self.log_text += s + "\n"

    def notify_order(self, order):
        if order.status in [order.Accepted]:
            # order already submitted/accepted - no action required
            return
        if order.status in [order.Submitted]:
            if order.isbuy():
                self.log(
                    f"Open: {self.data_open[0]:.3f}, Close: {self.data_close[0]:.3f}"
                )
                cost = order.created.size * order.created.price
                self.log(
                    "    "
                    f"BUY CREATED --- Size: {order.created.size:.3f}, "
                    f"Price: {order.created.price:.3f}, Cost: {cost:.3f}"
                )
            if order.issell():
                self.log(
                    f"Open: {self.data_open[0]:.3f}, Close: {self.data_close[0]:.3f}"
                )
                cost = order.created.size * order.created.price
                self.log(
                    "    "
                    f"SELL CREATED --- Size: {order.created.size:.3f}, "
                    f"Price: {order.created.price:.3f}, Cost: {cost:.3f}"
                )
            return

        # report executed order
        if order.status in [order.Completed]:
            if order.isbuy():
                cost = order.executed.size * order.executed.price
                self.log(
                    "    "
                    f"BUY EXECUTED --- Size: {order.executed.size:.3f}, "
                    f"Price: {order.executed.price:.3f}, "
                    f"Cost: {cost:.3f}, Commission: {order.executed.comm:.3f}"
                )
                self.price = order.executed.price
                self.comm = order.executed.comm
            else:
                cost = order.executed.size * order.executed.price
                self.log(
                    "    "
                    f"SELL EXECUTED --- Size: {order.executed.size:.3f}, "
                    f"Price: {order.executed.price:.3f}, "
                    f"Cost: {cost:.3f}, Commission: {order.executed.comm:.3f}"
                )
        elif order.status in [order.Canceled]:
            self.log("    " f"Order Canceled")
        elif order.status in [order.Rejected]:
            self.log("    " f"Order Rejected")
        elif order.status in [order.Margin]:
            self.log("    " f"Order Failed: Margin")

        # set no pending order
        self.order = None

    def notify_trade(self, trade):
        if not trade.isclosed:
            return
        self.log(
            "        "
            f" -> OPERATION RESULT --- Gross: {trade.pnl:.3f}, Net: {trade.pnlcomm:.3f}"
        )

    def next(self):
        # Check if we are in the market
        if not self.position:
            if self.data_predicted > 0:
                self.buy()
        else:
            # Already in the market ... we might sell
            if self.data_predicted < 0:
                self.sell()



"""
reporting with text
"""
import pandas as pd


def percentage_diff(df: pd.DataFrame, window=7) -> float:
    """
    Percentage difference given a window size.

    Arguments:
        df (:obj:`pandas.DataFrame`): dataframe of prices, must have
            a "Close" column.
        window (int): number of last n days used to computer percentage
            difference.
    """
    return (df.Close[-1] - df.Close[-window - 1]) / (df.Close[-window - 1]) * 100


def report_percentage_diff(
    df: pd.DataFrame, coin: str, windows=[1, 3, 7, 14, 60]
) -> str:
    """
    Produces verbose and pretty report on latest price
    difference from a given list of windows.

    Args:
        df (_type_): _description_
        coin (_type_): _description_
        windows (:obj:`list` of :obj:`int`, optional): list of windows
            to compute percentage difference. Defaults to [1, 3, 7, 14, 60].

    Returns:
        str: string report
    """
    s = f"**{coin}**\n"
    for window in windows:
        s = s + f"- {window}d: " + "{:.2f}".format(percentage_diff(df, window)) + " %\n"
    return s


def report_stock_gain(df: pd.DataFrame):
    """report stock gain

    Args:
        df (pd.DataFrame): dataframe with single day data with calculations

    Returns:
        str: string report
    """
    df = df[["Symbol", "Stock Gain / (Loss)"]]
    s = ""
    for idx, row in df.iterrows():
        s = (
            s
            + row["Symbol"]
            + " : "
            + "{:.2f}".format(row["Stock Gain / (Loss)"])
            + "\n"
        )
    return s


def report_coinbase_live_value(portfolio):
    """
    Nicely formatted report of accounts portfolio and user total
    balance in EUR.

    Args:
        portfolio (_type_): _descr_

    Returns:
        s (str): text
    """
    s = ""
    tot = 0
    for account in portfolio.coinbase.accounts:
        if float(account.native_balance.amount) > 0:
            s = s + str(account.currency) + " : " + str(account.native_balance) + "\n"
            tot += float(account.native_balance.amount)
    s = s + "---\n" + "Portfolio: EUR " + "{:.2f}".format(tot)
    return s


def report_tomorrow_prediction(binary_prediction: int):
    if binary_prediction > 0:
        return "BUY!"
    else:
        return "SELL / STAY OUT OF MARKET"
